"""
This contains common and (mostly) helpful utils.
"""

import inspect

from imviz.autoreload import ModuleReloader


class bundle(dict):
    """
    A dict, which allows dot notation access.
    """

    def __init__(self, *args, **kwargs):

        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self


STATICS = {}
"""
Contains all static function variables.
"""


def statics():
    """
    This (kind of) emulates the behavior of c++ static function variables.

    It returns a bundle, which is unique to the calling function.
    Like static variables the bundle is persisted between function calls.
    """

    caller = inspect.stack()[1]
    func_id = hash(caller.filename + caller.function)

    try:
        sts = STATICS[func_id]
    except KeyError:
        sts = bundle()
        STATICS[func_id] = sts

    return sts


RELOADER = ModuleReloader()
"""
Contains a global module reloader for easier access.
"""


def update_autoreload():
    """
    This checks all used python modules for changes (mtime), reloads
    the respective code, and updates existing code as far as possible.

    The "__main__" module cannot be reloaded due to python limitations.

    As this uses asynchronous modification scanning, the function
    must be called repeatedly in the main application loop.

    Returns True if at least one module was sucessfully reloaded.
    Returns False otherwise.
    """

    return RELOADER.reload()



"""
A module for automatic code reloading.

Caveats
=======
Reloading Python modules in a reliable way is in general difficult,
and unexpected things may occur. autoreload tries to work around
common pitfalls by replacing function code objects and parts of
classes previously in the module with new versions. This makes the
following things to work:
- Functions and classes imported via 'from xxx import foo' are upgraded
  to new versions when 'xxx' is reloaded.
- Methods and properties of classes are upgraded on reload, so that
  calling 'c.foo()' on an object 'c' created before the reload causes
  the new code for 'foo' to be executed.
Some of the known remaining caveats are:
- Replacing code objects does not always succeed: changing a @property
  in a class to an ordinary method or a method to a member variable
  can cause problems (but in old objects only).
- Functions that are removed (eg. via monkey-patching) from a module
  before it is reloaded are not upgraded.
- C extension modules cannot be reloaded, and so cannot be autoreloaded.
"""

# Copyright (C) 2000 Thomas Heller
# Copyright (C) 2008 Pauli Virtanen <pav@iki.fi>
# Copyright (C) 2012 The IPython Development Team
#
# The original IPython module was written by Pauli Virtanen, based on the
# autoreload code by Thomas Heller and distributed under BSD 3-Clause.
#
# The module was adapted for the imviz library.
#
# BSD 3-Clause License
#
# - Copyright (c) 2008-Present, IPython Development Team
# - Copyright (c) 2001-2007, Fernando Perez <fernando.perez@colorado.edu>
# - Copyright (c) 2001, Janko Hauser <jhauser@zscout.de>
# - Copyright (c) 2001, Nathaniel Gray <n8gray@caltech.edu>
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import gc
import sys
import types
import queue
import weakref
import traceback

from importlib import reload

import multiprocessing as mp


# ------------------------------------------------------------------------------
# Autoreload functionality
# ------------------------------------------------------------------------------


def scan_modules(requests, results):

    mtime_table = {}

    while True:

        req = requests.get()

        needs_reload = []

        for name, origin in req.items():

            if name in [None, "__mp_main__", "__main__"]:
                # we cannot reload(__main__) or reload(__mp_main__)
                continue

            if origin in [None, "built-in", "frozen"]:
                # builtins or frozen modules will likely not change
                continue

            try:
                mtime = os.stat(origin).st_mtime
            except OSError:
                continue

            try:
                if mtime_table[name] < mtime:
                    needs_reload.append(name)
            except KeyError:
                pass

            mtime_table[name] = mtime

        results.put(needs_reload)


class ModuleReloader:

    def __init__(self):

        self.waiting_for_scan = False

        self.scan_requests = mp.Queue(1)
        self.scan_results = mp.Queue(1)
        self.scan_process = mp.Process(
                target=scan_modules,
                args=[self.scan_requests, self.scan_results])
        self.scan_process.start()

        # (module-name, name) -> weakref, for replacing old code objects
        self.old_objects = {}

    def reload(self):
        """
        Check whether some modules need to be reloaded.
        """

        if not self.waiting_for_scan:

            # submit currently imported modules to scan process

            modules_to_scan = {}

            for name, mt in sys.modules.items():
                try:
                    modules_to_scan[name] = mt.__spec__.origin
                except AttributeError:
                    continue

            self.scan_requests.put(modules_to_scan)
            self.waiting_for_scan = True

            return False

        # check if the module scan was completed

        try:
            changed = self.scan_results.get_nowait()
            self.waiting_for_scan = False
        except queue.Empty:
            return False

        # ok scan is complete, reload changed modules

        if changed == []:
            return False

        for modname in changed:

            m = sys.modules.get(modname, None)

            try:
                superreload(m, reload, self.old_objects)
            except Exception:
                traceback.print_exc()

        return True


# ------------------------------------------------------------------------------
# superreload
# ------------------------------------------------------------------------------


func_attrs = [
    "__code__",
    "__defaults__",
    "__doc__",
    "__closure__",
    "__globals__",
    "__dict__",
]


def update_function(old, new):
    """Upgrade the code object of a function"""
    for name in func_attrs:
        try:
            setattr(old, name, getattr(new, name))
        except (AttributeError, TypeError):
            pass


def update_instances(old, new):
    """Use garbage collector to find all instances that refer to the old
    class definition and update their __class__ to point to the new class
    definition"""

    refs = gc.get_referrers(old)

    for ref in refs:
        if type(ref) is old:
            ref.__class__ = new


def update_class(old, new):
    """Replace stuff in the __dict__ of a class, and upgrade
    method code objects, and add new methods, if any"""
    for key in list(old.__dict__.keys()):
        old_obj = getattr(old, key)
        try:
            new_obj = getattr(new, key)
            # explicitly checking that comparison returns True to handle
            # cases where `==` doesn't return a boolean.
            if (old_obj == new_obj) is True:
                continue
        except AttributeError:
            # obsolete attribute: remove it
            try:
                delattr(old, key)
            except (AttributeError, TypeError):
                pass
            continue

        if update_generic(old_obj, new_obj):
            continue

        try:
            setattr(old, key, getattr(new, key))
        except (AttributeError, TypeError):
            pass  # skip non-writable attributes

    for key in list(new.__dict__.keys()):
        if key not in list(old.__dict__.keys()):
            try:
                setattr(old, key, getattr(new, key))
            except (AttributeError, TypeError):
                pass  # skip non-writable attributes

    # update all instances of class
    update_instances(old, new)


def update_property(old, new):
    """Replace get/set/del functions of a property"""
    update_generic(old.fdel, new.fdel)
    update_generic(old.fget, new.fget)
    update_generic(old.fset, new.fset)


def isinstance2(a, b, typ):
    return isinstance(a, typ) and isinstance(b, typ)


UPDATE_RULES = [
    (lambda a, b: isinstance2(a, b, type), update_class),
    (lambda a, b: isinstance2(a, b, types.FunctionType), update_function),
    (lambda a, b: isinstance2(a, b, property), update_property),
]
UPDATE_RULES.extend(
    [
        (
            lambda a, b: isinstance2(a, b, types.MethodType),
            lambda a, b: update_function(a.__func__, b.__func__),
        ),
    ]
)


def update_generic(a, b):
    for type_check, update in UPDATE_RULES:
        if type_check(a, b):
            update(a, b)
            return True
    return False


class StrongRef:
    def __init__(self, obj):
        self.obj = obj

    def __call__(self):
        return self.obj


mod_attrs = [
    "__name__",
    "__doc__",
    "__package__",
    "__loader__",
    "__spec__",
    "__file__",
    "__cached__",
    "__builtins__",
]


def append_obj(module, d, name, obj, autoload=False):
    in_module = hasattr(obj, "__module__") and obj.__module__ == module.__name__
    if autoload:
        # check needed for module global built-ins
        if not in_module and name in mod_attrs:
            return False
    else:
        if not in_module:
            return False

    key = (module.__name__, name)
    try:
        d.setdefault(key, []).append(weakref.ref(obj))
    except TypeError:
        pass
    return True


def superreload(module, reload=reload, old_objects=None, shell=None):
    """Enhanced version of the builtin reload function.
    superreload remembers objects previously in the module, and
    - upgrades the class dictionary of every old class in the module
    - upgrades the code object of every old function and method
    - clears the module's namespace before reloading
    """
    if old_objects is None:
        old_objects = {}

    # collect old objects in the module
    for name, obj in list(module.__dict__.items()):
        if not append_obj(module, old_objects, name, obj):
            continue
        key = (module.__name__, name)
        try:
            old_objects.setdefault(key, []).append(weakref.ref(obj))
        except TypeError:
            pass

    # reload module
    try:
        # clear namespace first from old cruft
        old_dict = module.__dict__.copy()
        old_name = module.__name__
        module.__dict__.clear()
        module.__dict__["__name__"] = old_name
        module.__dict__["__loader__"] = old_dict["__loader__"]
    except (TypeError, AttributeError, KeyError):
        pass

    try:
        module = reload(module)
    except:
        # restore module dictionary on failed reload
        module.__dict__.update(old_dict)
        raise

    # iterate over all objects and update functions & classes
    for name, new_obj in list(module.__dict__.items()):
        key = (module.__name__, name)
        if key not in old_objects:
            # here 'shell' acts both as a flag and as an output var
            if (
                shell is None
                or name == "Enum"
                or not append_obj(module, old_objects, name, new_obj, True)
            ):
                continue
            shell.user_ns[name] = new_obj

        new_refs = []
        for old_ref in old_objects[key]:
            old_obj = old_ref()
            if old_obj is None:
                continue
            new_refs.append(old_ref)
            update_generic(old_obj, new_obj)

        if new_refs:
            old_objects[key] = new_refs
        else:
            del old_objects[key]

    return module



"""
A very rudimentary and opinionated storage system.

The idea is not to store and load the entire object hierarchy.
Instead an already instantiated object hierarchy is update by the stored data.
This has the advantage that we can update most of the code without thinking
too much about how this will fit the already stored data.

The storage system also takes care of handling larger numpy arrays, which makes
working with images or points clouds much easier.
"""

import os
import json
import types
import hashlib
import importlib

import numpy as np


class Skip:
    """
    This signals that a value should not be serialized (only used internally).
    """
    pass


def full_type(o):
    """
    Returns the fully qualified type name of a given object.
    """

    cls = o.__class__
    module = cls.__module__

    if module == 'builtins':
        return cls.__qualname__

    return module + '.' + cls.__qualname__


def attrs_as_dict(obj):
    """
    Tries different methods to get a dict representation of obj.
    """

    attrs = {}

    if hasattr(obj, "__getstate__"):
        attrs = obj.__getstate__()
    elif hasattr(obj, "__dict__"):
        attrs = obj.__dict__
    elif isinstance(obj, dict):
        attrs = obj

    return attrs


def ext_setattr(obj, name, value):
    """
    Sets attributes objects and key-value pairs for dicts.
    """

    if type(obj) == dict:
        obj[name] = value
    else:
        setattr(obj, name, value)


class Serializer:
    """
    Serializes an object tree into a json file.
    Large numpy arrays are automatically referenced and stored externally.
    """

    def __init__(self, path, hide_private=True):

        self.path = path
        self.hide_private = hide_private

        self.ext_path = os.path.join(path, "extern")

        os.makedirs(self.ext_path, exist_ok=True)

        self.saved_arrays = set()

    def serialize(self, obj, name=""):

        if self.hide_private and len(name) > 0 and name[0] == "_":
            return Skip

        # skip any kind of function
        if (isinstance(obj, types.FunctionType)
                or isinstance(obj, types.MethodType)
                or isinstance(obj, types.LambdaType)):
            return Skip

        # special treatment for large numpy arrays
        if type(obj) == np.ndarray:
            if obj.size > 100:
                byte_view = obj.view(np.uint8)
                path = hashlib.sha1(byte_view).hexdigest() + ".npy"
                np.save(os.path.join(self.ext_path, path), obj)
                self.saved_arrays.add(path)
                return {
                    "__class__": "__extern__",
                    "path": path
                }
            else:
                return {
                    "__class__": full_type(obj),
                    "dtype": obj.dtype.name,
                    "data": obj.tolist()
                }

        # already saved and memory-mapped numpy arrays
        if type(obj) == np.memmap:
            return {
                "__class__": "__extern__",
                "path": os.path.basename(obj.filename)
            }

        if type(obj) == list:
            return [self.serialize(v) for v in obj]

        attrs = attrs_as_dict(obj)

        if attrs == {}:
            # in case we don't find any serializeable attributes
            # we assume we have a primitive type and return that
            return obj

        ser_attrs = {}

        for k, v in attrs.items():
            val = self.serialize(v, k)
            if val != Skip:
                ser_attrs[k] = val

        # store the full type so we can compare it later
        ser_attrs["__class__"] = full_type(obj)

        return ser_attrs


class Loader:
    """
    Loads an object tree from a json file.
    External numpy arrays are automatically dereferenced and memory mapped.
    """

    def __init__(self, path):

        self.path = path
        self.ext_path = os.path.join(path, "extern")

    def load(self, obj, json_obj):

        attr_names = attrs_as_dict(obj)

        for k, v in attr_names.items():

            if k not in json_obj:
                continue

            t = type(v)

            jv = json_obj[k]
            jt = type(jv)

            # special care for numpy arrays
            if jt == dict and "__class__" in jv:

                cls = jv["__class__"]

                if cls == "__extern__":
                    jv = np.load(os.path.join(self.ext_path, jv["path"]),
                                 mmap_mode="r")
                    # we are lying about this one (actually np.memmap)
                    # in practice memmap should behave just like ndarray
                    jt = np.ndarray
                elif cls == "numpy.ndarray":
                    jv = np.array(jv["data"], dtype=jv["dtype"])
                    jt = np.ndarray

            if t == dict and jt == dict:
                # recursion ...
                self.load(v, jv)
            elif t == list and jt == list:
                objs = []
                for jjv in jv:
                    jjt = type(jjv)
                    if jjt == dict and "__class__" in jjv:
                        mod_name, cls_name = jjv["__class__"].rsplit(".", 1)
                        try:
                            mod = importlib.import_module(mod_name)
                            cls = getattr(mod, cls_name)
                            # assumes default initializeable type
                            objs.append(cls())
                        except Exception as e:
                            print("Loading list element error:", e)
                    else:
                        objs.append(jjv)
                ext_setattr(obj, k, objs)
            elif t == jt:
                # this usually happens for primitive types
                ext_setattr(obj, k, jv)
            elif v is None:
                # we just use whatever is contained in json
                ext_setattr(obj, k, jv)
            else:
                # maybe just more recursion then?
                self.load(v, jv)


def save(obj, directory):
    """
    Stores obj under a given directory.
    The directory will be created if it not already exists.
    """

    # write to json

    os.makedirs(directory, exist_ok=True)

    ser = Serializer(directory)
    rep = ser.serialize(obj)

    state_path = os.path.join(directory, "state.json")

    with open(state_path, "w+") as fd:
        json.dump(rep, fd, indent=2)

    # remove unused external numpy arrays

    on_disk = set(os.listdir(ser.ext_path))
    unused = on_disk - ser.saved_arrays

    for f in unused:
        os.remove(os.path.join(ser.ext_path, f))


def load(obj, path):
    """
    Updates obj with data stored at the given path.
    """

    state_path = os.path.join(path, "state.json")

    if not os.path.exists(state_path):
        return

    with open(state_path, "r") as fd:
        json_state = json.load(fd)

    Loader(path).load(obj, json_state)
